// Sln2Makefile.cpp : Defines the entry point for the console application.
//

#include "stdafx.h"
#include "VcProjParser.hpp"
#include "Project.hpp"

using std::string;
using std::list;
using std::ifstream;
using std::ofstream;
using std::endl;
using std::cout;
using std::vector;


const static int	kBUFFER_SIZE			= 1024;

const static int	kMAX_ID_LENGTH			= 512;
const static int	kMAX_NAME_LENGTH		= 512;
const static int	kMAX_LOCATION_LENGTH	= 512;

void	fatal(const char *message, ...);

list<Project> projects;

string trim(string toTrim)
{
	if(toTrim.length() == 0)
		return toTrim;
	size_t b = toTrim.find_first_not_of(" \t");
	size_t e = toTrim.find_last_not_of(" \t");
	if(b == -1) // No non-spaces
		return "";
	return string(toTrim, b, e - b + 1);
}

Project ExtractHeaderFromString(string theString)
{
	Project temp;
	size_t quotePositions[8];

	size_t lastQuote = -1;
	for (int loop=0; loop<8; ++loop)
	{
		lastQuote = quotePositions[loop] = theString.find_first_of("\"",lastQuote+1);
	}
	temp.m_Id1 = theString.substr(quotePositions[0] + 1, quotePositions[1] - quotePositions[0] - 1);
	temp.m_Name = theString.substr(quotePositions[2] + 1, quotePositions[3] - quotePositions[2] - 1);
	temp.m_Location = theString.substr(quotePositions[4] + 1, quotePositions[5] - quotePositions[4] - 1);
	// convert '/' to '\'
	size_t pos = temp.m_Location.find("/");
	while ( pos != string::npos )
	{
		temp.m_Location.replace(pos, 1, "\\");
		pos = temp.m_Location.find("/");
	}
	temp.m_Id2 = theString.substr(quotePositions[6] + 1, quotePositions[7] - quotePositions[6] - 1);
	temp.m_TreeLevel = 0;
	temp.m_IsNeeded = false;

	return temp;
}

void RecursiveDependencyList(list<Project>::iterator whichProject)
{
	static int level=1;
	for (vector<Project::Dependency>::iterator i = whichProject->m_Dependencies.begin(); i != whichProject->m_Dependencies.end(); ++i)
	{
		for (list<Project>::iterator tempIt = projects.begin(); tempIt != projects.end(); ++tempIt)
		{
			if (tempIt->m_Id2.compare(i->m_Id2) == 0)
			{
				for (int loop=0; loop<level; ++loop)
					//cout << "\t";
				//cout << "Dependency : " << tempIt->m_Name << endl;
				if (level > tempIt->m_TreeLevel)
					tempIt->m_TreeLevel = level;
				tempIt->m_IsNeeded = true;
				++level;				
				RecursiveDependencyList(tempIt);
			}
		}
	}
	--level;
}

void PrintDepencies(const char *whichProject)
{
	for (list<Project>::iterator projHeadersIt = projects.begin(); projHeadersIt != projects.end(); ++projHeadersIt)
	{
		Project *pProjectHeader = &*projHeadersIt;
		if ( projHeadersIt->m_Name.compare(whichProject) == 0 )
		{
			//cout << "======================================================================" << endl;
			//cout << "Project : " << projHeadersIt->m_Name << endl;
			//cout << "Location: " << projHeadersIt->m_Location << endl;
			//cout << "======================================================================" << endl;
			
			pProjectHeader->m_IsNeeded = true;

			RecursiveDependencyList(projHeadersIt);

			for (vector<Project::Dependency>::iterator projDependIt = projHeadersIt->m_Dependencies.begin(); projDependIt != projHeadersIt->m_Dependencies.end(); ++projDependIt)
			{
				for (list<Project>::iterator tempIt = projects.begin(); tempIt != projects.end(); ++tempIt)
				{
					if (tempIt->m_Id2.compare(projDependIt->m_Id2) == 0)
					{
						////cout << "Dependency : " << tempIt->m_Name << endl;
						
					}
				}
			}
			
			//cout << endl;
			break;
		}
	}
}

void RecursivelyPrintLibsToBeLinked(const char *whichProject, ofstream *outfile)
{
	for (list<Project>::iterator projHeadersIt = projects.begin(); projHeadersIt != projects.end(); ++projHeadersIt)
	{
		Project *pProjectHeader = &*projHeadersIt;
		if ( projHeadersIt->m_Name.compare(whichProject) == 0 )
		{
			//cout << "======================================================================" << endl;
			//cout << "Project : " << projHeadersIt->m_Name << endl;
			//cout << "Location: " << projHeadersIt->m_Location << endl;
			//cout << "======================================================================" << endl;

			pProjectHeader->m_IsNeeded = true;

			for (vector<Project::Dependency>::iterator projDependIt = projHeadersIt->m_Dependencies.begin(); projDependIt != projHeadersIt->m_Dependencies.end(); ++projDependIt)
			{
				for (list<Project>::iterator tempIt = projects.begin(); tempIt != projects.end(); ++tempIt)
				{
					if (tempIt->m_Id2.compare(projDependIt->m_Id2) == 0)
					{
						RecursivelyPrintLibsToBeLinked(tempIt->m_Name.c_str(), outfile);
						// how many directories up is project base from here?
						int num=0;
						size_t pos=0;
						while (	pProjectHeader->m_Location.find("\\", pos) != -1 )
						{
							++num;
							pos = pProjectHeader->m_Location.find("\\", pos) + 1;
						}
						string projectDirectory = tempIt->m_Location.substr(0, tempIt->m_Location.find("\\"));
						*outfile << "LIBS += ";
						for (int i=0; i<num; ++i)
						{
							*outfile << "..\\";
						}
						*outfile << projectDirectory << "\\" << (pProjectHeader->m_IsDebug ? "PS2_EE_DEBUG\\" : pProjectHeader->m_IsRelease ? "PS2_EE_RELEASE\\" : "") << tempIt->m_Name << (pProjectHeader->m_IsDebug ? "_DEBUG.lib" : pProjectHeader->m_IsRelease ? "_RELEASE.lib" : "") << endl;						
					}
				}
			}

			//cout << endl;
			break;
		}
	}
}

int _tmain(int argc, _TCHAR* argv[])
{
	if (argc != 3)
	{
		string filename(argv[0]);
		printf("Syntax: %s Solution.sln Project\n", filename.substr(filename.find_last_of("\\") + 1).c_str() );
		printf("e.g. %s Pang.sln TextureManager\n", filename.substr(filename.find_last_of("\\") + 1).c_str() );
		return 0;
	}

	ifstream fSolution(argv[1]);
	char pszBuffer[kBUFFER_SIZE];

	if (!fSolution.is_open())
	{
		fatal("Could not open specified solution file (.sln)\n");
	}

	while ( !fSolution.eof() )
	{
		fSolution.getline(pszBuffer, kBUFFER_SIZE);
		string sBuffer(pszBuffer);
		
		size_t pos = sBuffer.find("Project(\"");
		if ( pos != -1 )
		{
			Project temp = ExtractHeaderFromString(sBuffer);
			
			fSolution.getline(pszBuffer, kBUFFER_SIZE);
			sBuffer = pszBuffer;
			if ( sBuffer.find("ProjectSection(ProjectDependencies)") != -1)
			{
				fSolution.getline(pszBuffer, kBUFFER_SIZE);
				sBuffer = pszBuffer;
				while ( sBuffer.find("EndProjectSection") == -1 )
				{
					Project::Dependency tempDependInfo;
					tempDependInfo.m_Id1 = sBuffer.substr(0, sBuffer.find_first_of("="));
					tempDependInfo.m_Id1 = trim(tempDependInfo.m_Id1);
					tempDependInfo.m_Id2 = sBuffer.substr(sBuffer.find_first_of("=") + 1);
					tempDependInfo.m_Id2 = trim(tempDependInfo.m_Id2);

					temp.m_Dependencies.push_back(tempDependInfo);
					
					fSolution.getline(pszBuffer, kBUFFER_SIZE);
					sBuffer = pszBuffer;
				}
			}

			temp.CalculateSourceAndObjectFiles();

			projects.push_back(temp);
		}
	}

	fSolution.close();

	// print out dependency tree for solution
	PrintDepencies(argv[2]);

	// print out list of source/object files for each project
	for (list<Project>::iterator tempIt = projects.begin(); tempIt != projects.end(); ++tempIt)
	{
		Project *pProjectHeader = &*tempIt;
		if (pProjectHeader->m_IsNeeded)
		{
			//cout << pProjectHeader->m_Name << endl;
			for (unsigned int i=0; i<pProjectHeader->m_Name.length(); ++i)
				;//cout << "=";

			//cout << endl;
			//cout << "C_SRCS =" << pProjectHeader->m_CSources << endl;
			//cout << "CXX_SRCS =" << pProjectHeader->m_CXXSources << endl;
			//cout << "VCL_SRCS =" << pProjectHeader->m_VclSources << endl;
			//cout << "OBJS =" << pProjectHeader->m_Objects<< endl;
			//cout << endl;
		}
	}

	// sort so dependencies are at beginning of list
	projects.sort();
	for (list<Project>::iterator tempIt = projects.begin(); tempIt != projects.end(); ++tempIt)
	{
		Project *pProjectHeader = &*tempIt;
		if (pProjectHeader->m_IsNeeded)
		{
			//cout << tempIt->m_TreeLevel << " : " << tempIt->m_Name << endl;
		}
	}

	//cout << endl;

	// create a global makefile which basically will just build the projects in correct order
	for (list<Project>::iterator tempIt = projects.begin(); tempIt != projects.end(); ++tempIt)
	{
		Project *pProjectHeader = &*tempIt;
		if (pProjectHeader->m_IsNeeded)
		{
			//cout << "cd " << pProjectHeader->m_Location.substr(0, pProjectHeader->m_Location.find_last_of("\\")) << endl;
			//cout << "make -f Makefile.txt" << endl;
			//cout << "cd .." << endl;
		}
		////cout << tempIt->m_TreeLevel << " : " << tempIt->m_Name << endl;
	}

	// write out this 'global' makefile
	ofstream finalMakefile("Makefile.txt");
	for (list<Project>::iterator tempIt = projects.begin(); tempIt != projects.end(); ++tempIt)
	{
		Project *pProjectHeader = &*tempIt;
		if (pProjectHeader->m_IsNeeded)
		{
			finalMakefile << "cd " << pProjectHeader->m_Location.substr(0, pProjectHeader->m_Location.find_last_of("\\")) << endl;
			finalMakefile << "make -f Makefile.txt" << endl;
			finalMakefile << "cd .." << endl;
		}
	}	
	finalMakefile.close();

	ifstream makefileHeader("c:\\cygwin\\usr\\local\\ps2\\makefile_header");	
	ifstream makefileFooter("c:\\cygwin\\usr\\local\\ps2\\makefile_footer");

	for (list<Project>::iterator tempIt = projects.begin(); tempIt != projects.end(); ++tempIt)
	{
		Project *pProjectHeader = &*tempIt;
		if (pProjectHeader->m_IsNeeded && pProjectHeader->m_Name == argv[2])
		{	
			makefileHeader.seekg(ifstream::beg);
			makefileFooter.seekg(ifstream::beg);

			string projectDirectory = pProjectHeader->m_Location.substr(0, pProjectHeader->m_Location.find("\\"));
			ofstream outfile((projectDirectory + "\\Makefile.txt").c_str());
			// make file header
			outfile << makefileHeader.rdbuf();
			outfile << endl << endl;
			// files to be built
			outfile << endl;
			outfile << "C_SRCS =" << pProjectHeader->m_CSources << endl;
			outfile << "CXX_SRCS =" << pProjectHeader->m_CXXSources << endl;
			outfile << "VCL_SRCS =" << pProjectHeader->m_VclSources << endl;
			outfile << "OBJS =" << pProjectHeader->m_Objects<< endl;
			outfile << endl;
			outfile << "vpath %.c " << pProjectHeader->m_CVPath << endl;
			outfile << "vpath %.cpp " << pProjectHeader->m_CPPVPath << endl;
			outfile << "vpath %.vcl " << pProjectHeader->m_VclVPath << endl;
			outfile << endl << endl;
			// add any libraries to be linked in...
			RecursivelyPrintLibsToBeLinked(argv[2], &outfile);
			RecursivelyPrintLibsToBeLinked(argv[2], &outfile);
			for (vector<Project::Dependency>::iterator projDependIt = tempIt->m_Dependencies.begin(); projDependIt != tempIt->m_Dependencies.end(); ++projDependIt)
			{
				for (list<Project>::iterator tempIt = projects.begin(); tempIt != projects.end(); ++tempIt)
				{
					if (tempIt->m_Id2.compare(projDependIt->m_Id2) == 0)
					{

					}
				}
			}
			
			outfile << makefileFooter.rdbuf();

			outfile.close();
		}
	}
	makefileFooter.close();
	makefileHeader.close();

	return 0;
}

void	fatal(const char *message, ...)
{
	//cout << message << endl;
}